# RFC3339 https://www.ietf.org/rfc/rfc3339.txt
import re
from datetime import datetime, timedelta, timezone
from typing import Optional

RFC_RE = re.compile(r"(\d{4})-(\d\d)-(\d\d)[ Tt](\d\d):(\d\d):(\d\d)([.,]\d+)?([zZ]|[-+](\d\d):?(\d\d)?)?")


class ParseError(ValueError):
    pass


def _tz(tzstr: str, hours: Optional[str], minutes: Optional[str]):
    if tzstr == "z" or tzstr == "Z":
        return timezone.utc
    # will raise TypeError if this are None
    h = int(hours)
    # minutes is optional
    m = int(minutes) if minutes is not None else 0
    # if not h and not m:
    #     return timezone.utc
    if h > 23 or m > 59:
        raise ParseError(f"Invalid timezone: {tzstr}")
    td = timedelta(hours=h, minutes=m)
    if tzstr.startswith("-"):
        td = -td
    return timezone(td)


def serialise(dt: datetime, *, assume_utc: bool = False):
    if dt.tzinfo is None:
        if assume_utc:
            dt = dt.replace(tzinfo=timezone.utc)
        else:
            raise ValueError("Datetime is not timezone aware")
    dt = dt.astimezone(timezone.utc)
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def deserialise(dtstr: str, *, assume_utc: bool = False):
    m = RFC_RE.match(dtstr)
    if not m:
        raise ParseError(f"Invalid RFC3339 string: {dtstr}")
    if m.endpos < len(dtstr) or m.endpos < len(dtstr.rstrip()):
        raise ParseError("Extra characters after end of datetime")
    year, month, day, hour, minute, second, frac, tzstr, tzhours, tzmins = m.groups()
    if tzstr is None:
        if assume_utc:
            tz = timezone.utc
        else:
            raise ParseError("RFC3339 string has no timezone part")
    else:
        tz = _tz(tzstr, tzhours, tzmins)
    return datetime(
        year=int(year),
        month=int(month),
        day=int(day),
        hour=int(hour),
        minute=int(minute),
        second=int(second),
        microsecond=int(frac[1:]) if frac is not None else 0,
        tzinfo=tz,
    )



from dataclasses import Field
from typing import Any, Protocol

from . import rfc3339 as _rfc3339


class Serialisation(Protocol):
    def serialise(self, value: Any) -> Any:
        ...

    def deserialise(self, into: Field, value: Any) -> Any:
        ...

    def as_list(self) -> "ListOf":
        return ListOf(self)

    def as_map(self) -> "MapOf":
        return MapOf(self)


class ListOf(Serialisation):
    def __init__(self, serialisation: Serialisation) -> None:
        self.serialisation = serialisation

    def serialise(self, value: Any) -> Any:
        ser = self.serialisation.serialise
        return [ser(i) for i in value]

    def deserialise(self, into: Field, value: Any) -> Any:
        des = self.serialisation.deserialise
        return [des(into, i) for i in value]


class MapOf(Serialisation):
    def __init__(self, serialisation: Serialisation) -> None:
        self.serialisation = serialisation

    def serialise(self, value: Any) -> Any:
        ser = self.serialisation.serialise
        return {k: ser(v) for k, v in value.items()}

    def deserialise(self, into: Field, value: Any) -> Any:
        des = self.serialisation.deserialise
        return {k: des(into, v) for k, v in value.items()}


class Identity(Serialisation):
    def serialise(self, value: Any) -> Any:
        return value

    def deserialise(self, into: Field, value: Any) -> Any:
        return value


class RFC3339Enc(Serialisation):
    def serialise(self, value: Any) -> Any:
        return _rfc3339.serialise(value) if value is not None else None

    def deserialise(self, into: Field, value: Any) -> Any:
        return _rfc3339.deserialise(value)


class ObjEnc(Serialisation):
    def __init__(self, cls: type) -> None:
        self._cls = cls

    def serialise(self, value: Any) -> Any:
        return value.to_json() if value is not None else None

    def deserialise(self, into: Field, value: Any) -> Any:
        return self._cls.from_json(value)


RFC3339 = RFC3339Enc()



from dataclasses import MISSING as DC_MISSING
from dataclasses import Field, InitVar, is_dataclass

try:
    from types import NoneType
except ImportError:
    NoneType = type(None)
from typing import Any, Callable, ClassVar, List, Optional, Tuple, Type, TypeVar, Union, get_args

from typing_extensions import Annotated, get_origin

from .attributes import JSON, MISSING

A = TypeVar("A")


class MissingJSONDefinitionError(LookupError):
    pass


def is_missing(value: Any) -> bool:
    return value is DC_MISSING


def has_default(value: Field) -> bool:
    return not is_missing(value.default) or not is_missing(value.default_factory)


def get_json_attribute(typ: Any) -> Optional[JSON]:
    if get_origin(typ) is Annotated:
        for param in get_args(typ)[1:]:
            if isinstance(param, JSON):
                return param
    return None


def is_nullable(typ: Any) -> bool:
    origin = get_origin(typ)
    if origin is Union:
        return NoneType in get_args(typ)
    if origin is Annotated:
        return is_nullable(get_args(typ)[0])
    return False


def create_builder(json: List[Tuple[Field, JSON]]) -> Callable[[Any], A]:
    def from_json(cls: Type[A], input_: Any) -> A:
        args = {f.name: v for f, v in ((f, j.extract(input_, None)) for f, j in json) if v is not MISSING}
        return cls(**args)

    return from_json


def create_writer(json: List[Tuple[Field, JSON]]) -> Callable[[Any], Any]:
    def to_json(inst: Any) -> Any:
        r = {}
        for f, j in json:
            r.update(j.prepare(getattr(inst, f.name)))
        return r

    return to_json


def validate(cls: type, *, generate_default_defn: bool = False) -> None:
    if not is_dataclass(cls):
        raise TypeError("Class must be a dataclass")

    cls_builders = []

    for field in cls.__dataclass_fields__.values():
        json = validate_field(field, generate_default_defn=generate_default_defn)
        if json is not None:
            cls_builders.append((field, json))

    return cls_builders


def validate_field(field: Field, generate_default_defn: bool = False) -> Optional[JSON]:
    typ = field.type
    origin = get_origin(typ)
    json = get_json_attribute(typ)

    if origin is ClassVar:
        if json is not None:
            raise TypeError("JSON attributes found on class variable.")
        return
    if not field.init:
        return
    if origin is InitVar:
        # init-only, if this has no defaults we can't create from json
        if not has_default(field):
            raise TypeError(
                f"Init-only field {field.name} has no default, cannot generate from_json() method. if this is optional, please set a default."
            )
        # TODO: support initvars from json
        return
    if json is None:
        if generate_default_defn:
            json = JSON(field.name, nullable=is_nullable(typ))
        else:
            if not has_default(field):
                raise TypeError(
                    f"Field {field.name} has no default, cannot generate from_json() method. if this is optional, please set a default."
                )
            return
    if json.nullable and not is_nullable(typ):
        raise TypeError(f"Field {field.name} is nullable but type is not Optional[...]")
    if json.omit_empty and not has_default(field):
        raise TypeError(f"Field {field.name} is omit_empty but has no default")

    return json


def json_schema(
    cls: Optional[type] = None, generate_default_defn: bool = False, from_json: bool = True, to_json: bool = True
):
    def decorator(cls: type) -> type:
        builders = validate(cls, generate_default_defn=generate_default_defn)
        if from_json and not hasattr(cls, "from_json"):
            cls.from_json = classmethod(create_builder(builders))
        if to_json and not hasattr(cls, "to_json"):
            cls.to_json = create_writer(builders)
        return cls

    if cls is None:
        return decorator
    return decorator(cls)



from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

from .serialisation import Serialisation


@(lambda cls: cls())
class MISSING:
    __slots__ = ()

    def __bool__(self):
        return False

    def __repr__(self) -> str:
        return "<MISSING>"


@dataclass
class JSON:
    name: str
    omit_empty: bool = False
    nullable: bool = False
    serialisation: Optional[Serialisation] = None
    nonempty_predicate: Optional[Callable[[Any], bool]] = None
    omitted_value: Any = MISSING

    def _nonempty(self, value: Any):
        nep = self.nonempty_predicate
        if nep is None:
            return value is not None
        return nep(value)

    def serialise(self, value: Any) -> Any:
        if value is None:
            if self.nullable:
                return None
            raise TypeError("Got 'None' but field is not nullable")
        return value if self.serialisation is None else self.serialisation.serialise(value)

    def deserialise(self, value: Any) -> Any:
        return value if self.serialisation is None else self.serialisation.deserialise(None, value)

    @staticmethod
    def default(key: str, value: Any) -> Dict[str, Any]:
        return {key: value}

    def prepare(self, value: Any) -> Dict[str, Any]:
        if not self.omit_empty or self._nonempty(value):
            return {self.name: self.serialise(value)}
        return {}

    def extract(self, input_: Any, k) -> Any:
        try:
            value = input_[self.name]
        except LookupError:
            if self.omit_empty:
                return self.omitted_value
            raise
        else:
            if value is None and not self.nullable:
                raise TypeError("Got 'None' but field is not nullable")
            return self.deserialise(value)
